{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE InstanceSigs #-}
{-# OPTIONS_GHC -Wno-unused-top-binds #-}

module ApplicativeChapter () where

import Data.Function (const)
import GHC.Base ( id, Functor(..) )
import Data.Maybe (Maybe(..))
import Data.Function (($))
import Data.List (zipWith)

liftA2 :: ApplicativeChapter.Applicative f => (a -> b -> c) -> f a -> f b -> f c
liftA2 f x = (<*>) (fmap f x)

class Functor f => Applicative f where
  pure :: a -> f a
  infixl 4 <*>, *>
  (<*>) :: f (a -> b) -> f a -> f b

  (*>) :: f a -> f b -> f b
  a1 *> a2 = (id <$ a1) <*> a2

  (<*) :: f a -> f b -> f a
  (<*) = liftA2 const

-- ??

-- Maybe is Applicative

instance Applicative Maybe where
  (<*>) :: Maybe (a -> b) -> Maybe a -> Maybe b
  (<*>) Nothing _ = Nothing
  (<*>) (Just f) x = fmap f x

  pure :: a -> Maybe a
  pure = Just

-- List can be Applicative in 2 different ways

newtype ZipList a = ZipList { getZipList :: [a] }

instance Applicative ZipList where
  pure :: a -> ZipList a
  pure a = ZipList [a]

  (<*>) :: ZipList (a -> b) -> ZipList a -> ZipList b
  (ZipList gs) <*> (ZipList xs) = ZipList (zipWith ($) gs xs)
